{
  PasWebP
  Free WebP library for Free Pascal

  Headers translated to Object Pascal and C code precompliled
  by SileNTViP
  for Vampyre Imaging Library (http://imaginglib.sourceforge.net).

  Supported compilers: Free Pascal
  Supported platforms (tested): Windows 32bit

  WebP codec Homepage: https://developers.google.com/speed/webp/
  Source Code: https://chromium.googlesource.com/webm/libwebp/
  Compiled libs: https://storage.googleapis.com/downloads.webmproject.org/releases/webp/index.html

  Current Version: 1.00 (WebP codec 1.0.0)

  History:
    v1.00
      - release
}

unit PasWebP;

{$IFDEF FPC}
  { Free Pascal settings }
  {$PACKRECORDS 8}
  {$PACKENUM 4}
{$ELSE}
  { Delphi settings }
  {$DEFINE DCC}
  {$ALIGN 8}
  {$MINENUMSIZE 4}
{$ENDIF}

interface

{$IFDEF FPC}
uses
  Math;
{$ENDIF}

{ From webp\decode.h }

type
  WebPIDecoder = record end;
  PWebPIDecoder = ^WebPIDecoder;

const
  WEBP_DECODER_ABI_VERSION = $0208;

{ Return the decoder's version number, packed in hexadecimal using 8bits for
  each of major/minor/revision. E.g: v2.5.7 is 0x020507.}
function WebPGetDecoderVersion: Integer; cdecl; external;

{ Retrieve basic header information: width, height.
  This function will also validate the header, returning true on success,
  false otherwise. '*width' and '*height' are only valid on successful return.
  Pointers 'width' and 'height' can be passed NULL if deemed irrelevant.}
function WebPGetInfo(const Data: PByte; Data_Size: SIZE_T; Width: PInteger;
  Height: PInteger): Integer; cdecl; external;

{ Decodes WebP images pointed to by 'data' and returns RGBA samples, along
  with the dimensions in *width and *height. The ordering of samples in
  memory is R, G, B, A, R, G, B, A... in scan order (endian-independent).
  The returned pointer should be deleted calling WebPFree().
  Returns NULL in case of error.}
function WebPDecodeRGBA(const Data: PByte; Data_Size: SIZE_T; Width: PInteger;
  Height: PInteger): PByte; cdecl; external;

{ Same as WebPDecodeRGBA, but returning A, R, G, B, A, R, G, B... ordered data.}
function WebPDecodeARGB(const Data: PByte; Data_Size: SIZE_T; Width: PInteger;
  Height: PInteger): PByte; cdecl; external;

{ Same as WebPDecodeRGBA, but returning B, G, R, A, B, G, R, A... ordered data.}
function WebPDecodeBGRA(const Data: PByte; Data_Size: SIZE_T; Width: PInteger;
  Height: PInteger): PByte; cdecl; external;

{ Same as WebPDecodeRGBA, but returning R, G, B, R, G, B... ordered data.
  If the bitstream contains transparency, it is ignored.}
function WebPDecodeRGB(const Data: PByte; Data_Size: SIZE_T; Width: PInteger;
  Height: PInteger): PByte; cdecl; external;

{ Same as WebPDecodeRGB, but returning B, G, R, B, G, R... ordered data.}
function WebPDecodeBGR(const Data: PByte; Data_Size: SIZE_T; Width: PInteger;
  Height: PInteger): PByte; cdecl; external;

{ Decode WebP images pointed to by 'data' to Y'UV format(*). The pointer
  returned is the Y samples buffer. Upon return, *u and *v will point to
  the U and V chroma data. These U and V buffers need NOT be passed to
  WebPFree(), unlike the returned Y luma one. The dimension of the U and V
  planes are both (*width + 1) / 2 and (*height + 1)/ 2.
  Upon return, the Y buffer has a stride returned as '*stride', while U and V
  have a common stride returned as '*uv_stride'.
  Return NULL in case of error.
  (*) Also named Y'CbCr. See: http://en.wikipedia.org/wiki/YCbCr}
function WebPDecodeYUV(const Data: PByte; Data_Size: SIZE_T; Width: PInteger;
  Height: PInteger; U: PPByte; V: PPByte; Stride: PInteger; UV_Stride: PInteger):
  PByte; cdecl; external;

{ Releases memory returned by the WebPDecode*() functions above.}
procedure WebPFree(Ptr: Pointer); cdecl; external;

{ These five functions are variants of the above ones, that decode the image
  directly into a pre-allocated buffer 'output_buffer'. The maximum storage
  available in this buffer is indicated by 'output_buffer_size'. If this
  storage is not sufficient (or an error occurred), NULL is returned.
  Otherwise, output_buffer is returned, for convenience.
  The parameter 'output_stride' specifies the distance (in bytes)
  between scanlines. Hence, output_buffer_size is expected to be at least
  output_stride x picture-height.}
function WebPDecodeRGBAInto(
  const Data: PByte; Data_Size: SIZE_T; Output_Buffer: PByte;
  Output_Buffer_Size: SIZE_T; Output_Stride: Integer): PByte; cdecl; external;
function WebPDecodeARGBInto(
  const Data: PByte; Data_Size: SIZE_T; Output_Buffer: PByte;
  Output_Buffer_Size: SIZE_T; Output_Stride: Integer): PByte; cdecl; external;
function WebPDecodeBGRAInto(
  const Data: PByte; Data_Size: SIZE_T; Output_Buffer: PByte;
  OutputBufferSize: SIZE_T; Output_Stride: Integer): PByte; cdecl; external;

{ RGB and BGR variants. Here too the transparency information, if present,
  will be dropped and ignored.}
function WebPDecodeRGBInto(
  const Data: PByte; Data_Size: SIZE_T; Output_Buffer: PByte;
  Output_Buffer_Size: SIZE_T; Output_Stride: Integer): PByte; cdecl; external;
function WebPDecodeBGRInto(
  const Data: PByte; Data_Size: SIZE_T; Output_Buffer: PByte;
  Output_Buffer_Size: SIZE_T; Output_Stride: Integer): PByte; cdecl; external;

{ WebPDecodeYUVInto() is a variant of WebPDecodeYUV() that operates directly
  into pre-allocated luma/chroma plane buffers. This function requires the
  strides to be passed: one for the luma plane and one for each of the
  chroma ones. The size of each plane buffer is passed as 'luma_size',
  'u_size' and 'v_size' respectively.
  Pointer to the luma plane ('*luma') is returned or NULL if an error occurred
  during decoding (or because some buffers were found to be too small).}
function WebPDecodeYUVInto(
  const Data: PByte; Data_Size: SIZE_T;
  Luma: PByte; Luma_Size: SIZE_T; Luma_Stride: Integer;
  U: PByte; U_Size: SIZE_T; U_Stride: Integer;
  V: PByte; V_Size: SIZE_T; V_Stride: Integer): PByte; cdecl; external;

//------------------------------------------------------------------------------
// Output colorspaces and buffer

// Colorspaces
// Note: the naming describes the byte-ordering of packed samples in memory.
// For instance, MODE_BGRA relates to samples ordered as B,G,R,A,B,G,R,A,...
// Non-capital names (e.g.:MODE_Argb) relates to pre-multiplied RGB channels.
// RGBA-4444 and RGB-565 colorspaces are represented by following byte-order:
// RGBA-4444: [r3 r2 r1 r0 g3 g2 g1 g0], [b3 b2 b1 b0 a3 a2 a1 a0], ...
// RGB-565: [r4 r3 r2 r1 r0 g5 g4 g3], [g2 g1 g0 b4 b3 b2 b1 b0], ...
// In the case WEBP_SWAP_16BITS_CSP is defined, the bytes are swapped for
// these two modes:
// RGBA-4444: [b3 b2 b1 b0 a3 a2 a1 a0], [r3 r2 r1 r0 g3 g2 g1 g0], ...
// RGB-565: [g2 g1 g0 b4 b3 b2 b1 b0], [r4 r3 r2 r1 r0 g5 g4 g3], ...

type
  WEBP_CSP_MODE = (
    MODE_RGB = 0, MODE_RGBA = 1,
    MODE_BGR = 2, MODE_BGRA = 3,
    MODE_ARGB = 4, MODE_RGBA_4444 = 5,
    MODE_RGB_565 = 6,
    // RGB-premultiplied transparent modes (alpha value is preserved)
    MODEP_rgbA = 7,
    MODEP_bgrA = 8,
    MODEP_Argb = 9,
    MODEP_rgbA_4444 = 10,
    // YUV modes must come after RGB ones.
    MODE_YUV = 11, MODE_YUVA = 12,  // yuv 4:2:0
    MODE_LAST = 13
  );

//------------------------------------------------------------------------------
// WebPDecBuffer: Generic structure for describing the output sample buffer.

type
  WebPRGBABuffer = record // view as RGBA
    RGBA: PByte;          // pointer to RGBA samples
    Stride: Integer;      // stride in bytes from one scanline to the next.
    Size: SIZE_T;         // total size of the *rgba buffer.
  end;

  WebPYUVABuffer = record // view as YUVA
    Y: PByte;                    // pointer to luma, chroma U/V, alpha samples
    U, V, A: PPByte;
    Y_Stride: Integer;            // luma stride
    U_Stride, V_Stride: Integer;   // chroma strides
    A_Stride: Integer;            // alpha stride
    Y_Size: SIZE_T;               // luma plane size
    U_Size, V_Size: SIZE_T;        // chroma planes size
    A_Size: SIZE_T;               // alpha-plane size
  end;

  // Output buffer
  WebPDecBuffer = record
    ColorSpace: WEBP_CSP_MODE; // Colorspace.
    Width, Height: Integer;    // Dimensions.
    Is_External_Memory: Integer; // If non-zero, 'internal_memory' pointer is not
                               // used. If value is '2' or more, the external
                               // memory is considered 'slow' and multiple
                               // read/write will be avoided.
    U: record                  // Nameless union of buffer parameters.
      RGBA: WebPRGBABuffer;
      YUVA: WebPYUVABuffer;
    end;
    Pad: array [0..3] of Cardinal; // padding for later use
    Private_Memory: PByte;     // Internally allocated memory (only when
                               // is_external_memory is 0). Should not be used
                               // externally, but accessed via the buffer union.
  end;
  PWebPDecBuffer = ^WebPDecBuffer;

{ Internal, version-checked, entry point}
function WebPInitDecBufferInternal(Buffer: PWebPDecBuffer; Version: Integer):
  Integer; cdecl; external;

{ Initialize the structure as empty. Must be called before any other use.
  Returns false in case of version mismatch}
function WebPInitDecBuffer(Buffer: PWebPDecBuffer): Integer;

{ Free any memory associated with the buffer. Must always be called last.
  Note: doesn't free the 'buffer' structure itself.}
procedure WebPFreeDecBuffer(Buffer: PWebPDecBuffer); cdecl; external;

//------------------------------------------------------------------------------
// Enumeration of the status codes

type
  VP8StatusCode = (
    VP8_STATUS_OK = 0,
    VP8_STATUS_OUT_OF_MEMORY,
    VP8_STATUS_INVALID_PARAM,
    VP8_STATUS_BITSTREAM_ERROR,
    VP8_STATUS_UNSUPPORTED_FEATURE,
    VP8_STATUS_SUSPENDED,
    VP8_STATUS_USER_ABORT,
    VP8_STATUS_NOT_ENOUGH_DATA
  );

//------------------------------------------------------------------------------
// Incremental decoding
{
   This API allows streamlined decoding of partial data.
   Picture can be incrementally decoded as data become available thanks to the
   WebPIDecoder object. This object can be left in a SUSPENDED state if the
   picture is only partially decoded, pending additional input.
   Code example:

   var
     idec: PWebPIDecoder;
   ...
   WebPInitDecBuffer(@output_buffer);
   output_buffer.colorspace = mode;
   ...
   idec := WebPINewDecoder(@output_buffer);
   while additional_data_is_available do begin
     // ... (get additional data in some new_data[] buffer)
     status := WebPIAppend(idec, new_data, new_data_size);
     if (status <> VP8_STATUS_OK) and (status <> VP8_STATUS_SUSPENDED) then
       Break;    // an error occurred.

     // The above call decodes the current available buffer.
     // Part of the image can now be refreshed by calling
     // WebPIDecGetRGB()/WebPIDecGetYUVA() etc.
   end;
   WebPIDelete(idec);
}

{ Creates a new incremental decoder with the supplied buffer parameter.
  This output_buffer can be passed NULL, in which case a default output buffer
  is used (with MODE_RGB). Otherwise, an internal reference to 'output_buffer'
  is kept, which means that the lifespan of 'output_buffer' must be larger than
  that of the returned WebPIDecoder object.
  The supplied 'output_buffer' content MUST NOT be changed between calls to
  WebPIAppend() or WebPIUpdate() unless 'output_buffer.is_external_memory' is
  not set to 0. In such a case, it is allowed to modify the pointers, size and
  stride of output_buffer.u.RGBA or output_buffer.u.YUVA, provided they remain
  within valid bounds.
  All other fields of WebPDecBuffer MUST remain constant between calls.
  Returns NULL if the allocation failed.}
function WebPINewDecoder(Output_Buffer: PWebPDecBuffer):
  PWebPIDecoder; cdecl; external;

{ This function allocates and initializes an incremental-decoder object, which
  will output the RGB/A samples specified by 'csp' into a preallocated
  buffer 'output_buffer'. The size of this buffer is at least
  'output_buffer_size' and the stride (distance in bytes between two scanlines)
  is specified by 'output_stride'.
  Additionally, output_buffer can be passed NULL in which case the output
  buffer will be allocated automatically when the decoding starts. The
  colorspace 'csp' is taken into account for allocating this buffer. All other
  parameters are ignored.
  Returns NULL if the allocation failed, or if some parameters are invalid.}
function WebPINewRGB(CSP: WEBP_CSP_MODE; Output_Buffer: PByte;
  Output_Buffer_Size: SIZE_T; Output_Stride: Integer): PWebPIDecoder; cdecl; external;

{ This function allocates and initializes an incremental-decoder object, which
  will output the raw luma/chroma samples into a preallocated planes if
  supplied. The luma plane is specified by its pointer 'luma', its size
  'luma_size' and its stride 'luma_stride'. Similarly, the chroma-u plane
  is specified by the 'u', 'u_size' and 'u_stride' parameters, and the chroma-v
  plane by 'v' and 'v_size'. And same for the alpha-plane. The 'a' pointer
  can be pass NULL in case one is not interested in the transparency plane.
  Conversely, 'luma' can be passed NULL if no preallocated planes are supplied.
  In this case, the output buffer will be automatically allocated (using
  MODE_YUVA) when decoding starts. All parameters are then ignored.
  Returns NULL if the allocation failed or if a parameter is invalid.}
function WebPINewYUVA(
  Luma: PByte; Luma_Size: SIZE_T; Luma_Stride: Integer;
  U: PByte; U_Size: SIZE_T; U_Stride: Integer;
  V: PByte; V_Size: SIZE_T; V_Stride: Integer;
  A: PByte; A_Size: SIZE_T; A_Stride: Integer): PWebPIDecoder; cdecl; external;

{ Deprecated version of the above, without the alpha plane.
  Kept for backward compatibility.}
function WebPINewYUV(
  Luma: PByte; Luma_Size: SIZE_T; Luma_Stride: Integer;
  U: PByte; U_Size: SIZE_T; U_Stride: Integer;
  V: PByte; V_Size: SIZE_T; V_Stride: Integer): PWebPIDecoder; cdecl; external;

{ Deletes the WebPIDecoder object and associated memory. Must always be called
  if WebPINewDecoder, WebPINewRGB or WebPINewYUV succeeded.}
procedure WebPIDelete(IDec: PWebPIDecoder); cdecl; external;

{ Copies and decodes the next available data. Returns VP8_STATUS_OK when
  the image is successfully decoded. Returns VP8_STATUS_SUSPENDED when more
  data is expected. Returns error in other cases.}
function WebPIAppend(IDec: PWebPIDecoder; const Data: PByte; Data_Size: SIZE_T):
  VP8StatusCode; cdecl; external;

{ A variant of the above function to be used when data buffer contains
  partial data from the beginning. In this case data buffer is not copied
  to the internal memory.
  Note that the value of the 'data' pointer can change between calls to
  WebPIUpdate, for instance when the data buffer is resized to fit larger data.}
function WebPIUpdate(IDec: PWebPIDecoder; const Data: PByte; Data_Size: SIZE_T):
  VP8StatusCode; cdecl; external;

{ Returns the RGB/A image decoded so far. Returns NULL if output params
  are not initialized yet. The RGB/A output type corresponds to the colorspace
  specified during call to WebPINewDecoder() or WebPINewRGB().
  *last_y is the index of last decoded row in raster scan order. Some pointers
  (*last_y, *width etc.) can be NULL if corresponding information is not
  needed. The values in these pointers are only valid on successful (non-NULL)
  return.}
function WebPIDecGetRGB(const IDec: PWebPIDecoder; Last_Y: PInteger;
  Width: PInteger; Height: PInteger; Stride: PInteger): PByte; cdecl; external;

{ Same as above function to get a YUVA image. Returns pointer to the luma
  plane or NULL in case of error. If there is no alpha information
  the alpha pointer '*a' will be returned NULL.}
function WebPIDecGetYUVA( const IDec: PWebPIDecoder; Last_Y: PInteger;
  U: PPByte; V: PPByte; A: PPByte;
  Width: PInteger; Height: Pinteger;
  Stride: PInteger; UV_Stride: PInteger; A_Stride: PInteger):
  PByte; cdecl; external;

{ Deprecated alpha-less version of WebPIDecGetYUVA(): it will ignore the
  alpha information (if present). Kept for backward compatibility.}
function WebPIDecGetYUV(const IDec: PWebPIDecoder; Last_Y:
  PInteger; U: PPByte; V: PPByte; Width: PInteger; Height: PInteger;
  Stride: PInteger; UV_Stride: PInteger): PByte;

{ Generic call to retrieve information about the displayable area.
  If non NULL, the left/right/width/height pointers are filled with the visible
  rectangular area so far.
  Returns NULL in case the incremental decoder object is in an invalid state.
  Otherwise returns the pointer to the internal representation. This structure
  is read-only, tied to WebPIDecoder's lifespan and should not be modified.}
function WebPIDecodedArea(const IDec: PWebPIDecoder;
  Left: PInteger; Top: PInteger; Width: PInteger; Height: PInteger):
  PWebPDecBuffer; cdecl; external;

//------------------------------------------------------------------------------
// Advanced decoding parametrization
//
//  Code sample for using the advanced decoding API
{
     // A) Init a configuration object
     WebPDecoderConfig config;
     CHECK(WebPInitDecoderConfig(&config));

     // B) optional: retrieve the bitstream's features.
     CHECK(WebPGetFeatures(data, data_size, &config.input) == VP8_STATUS_OK);

     // C) Adjust 'config', if needed
     config.no_fancy_upsampling = 1;
     config.output.colorspace = MODE_BGRA;
     // etc.

     // Note that you can also make config.output point to an externally
     // supplied memory buffer, provided it's big enough to store the decoded
     // picture. Otherwise, config.output will just be used to allocate memory
     // and store the decoded picture.

     // D) Decode!
     CHECK(WebPDecode(data, data_size, &config) == VP8_STATUS_OK);

     // E) Decoded image is now in config.output (and config.output.u.RGBA)

     // F) Reclaim memory allocated in config's object. It's safe to call
     // this function even if the memory is external and wasn't allocated
     // by WebPDecode().
     WebPFreeDecBuffer(&config.output);
}

type
  // Features gathered from the bitstream
  WebPBitstreamFeatures = record
    Width: Integer;          // Width in pixels, as read from the bitstream.
    Height: Integer;         // Height in pixels, as read from the bitstream.
    Has_Alpha: Integer;      // True if the bitstream contains an alpha channel.
    Has_Animation: Integer;  // True if the bitstream is an animation.
    Format: Integer;         // 0 = undefined (/mixed), 1 = lossy, 2 = lossless

    Pad: array [0..4] of Cardinal;    // padding for later use
  end;
  PWebPBitstreamFeatures = ^WebPBitstreamFeatures;

{ Internal, version-checked, entry point}
function WebPGetFeaturesInternal(const Data: PByte; Data_Size: SIZE_T;
  Features: PWebPBitstreamFeatures;
  Version: Integer): VP8StatusCode; cdecl; external;

{ Retrieve features from the bitstream. The *features structure is filled
  with information gathered from the bitstream.
  Returns VP8_STATUS_OK when the features are successfully retrieved. Returns
  VP8_STATUS_NOT_ENOUGH_DATA when more data is needed to retrieve the
  features from headers. Returns error in other cases.}
function WebPGetFeatures(const Data: PByte; Data_Size: SIZE_T;
  Features: PWebPBitstreamFeatures): VP8StatusCode;

type
  // Decoding options
  WebPDecoderOptions = record
    Bypass_Filtering: Integer;             // if true, skip the in-loop filtering
    No_Fancy_UpSampling: Integer;          // if true, use faster pointwise upsampler
    Use_Cropping: Integer;                 // if true, cropping is applied _first_
    Crop_Left, Crop_Top: Integer;          // top-left position for cropping.
                                           // Will be snapped to even values.
    Crop_Width, Crop_Height: Integer;      // dimension of the cropping area
    Use_Scaling: Integer;                  // if true, scaling is applied _afterward_
    Scaled_Width, Scaled_Height: Integer;  // final resolution
    Use_Threads: Integer;                  // if true, use multi-threaded decoding
    Dithering_Strength: Integer;           // dithering strength (0=Off, 100=full)
    Flip: Integer;                         // flip output vertically
    Alpha_Dithering_Strength: Integer;     // alpha dithering strength in [0..100]

    Pad: array [0..4] of Cardinal;         // padding for later use
  end;

  // Main object storing the configuration for advanced decoding.
  WebPDecoderConfig = record
    Input: WebPBitstreamFeatures;  // Immutable bitstream features (optional)
    Output: WebPDecBuffer;         // Output buffer (can point to external mem)
    Options: WebPDecoderOptions;   // Decoding options
  end;
  PWebPDecoderConfig = ^WebPDecoderConfig;

{ Internal, version-checked, entry point}
function WebPInitDecoderConfigInternal(Config: PWebPDecoderConfig;
  Version: Integer): Integer; cdecl; external;

{ Initialize the configuration as empty. This function must always be
  called first, unless WebPGetFeatures() is to be called.
  Returns false in case of mismatched version.}
function WebPInitDecoderConfig(Config: PWebPDecoderConfig): Integer;

{ Instantiate a new incremental decoder object with the requested
  configuration. The bitstream can be passed using 'data' and 'data_size'
  parameter, in which case the features will be parsed and stored into
  config->input. Otherwise, 'data' can be NULL and no parsing will occur.
  Note that 'config' can be NULL too, in which case a default configuration
  is used. If 'config' is not NULL, it must outlive the WebPIDecoder object
  as some references to its fields will be used. No internal copy of 'config'
  is made.
  The return WebPIDecoder object must always be deleted calling WebPIDelete().
  Returns NULL in case of error (and config->status will then reflect
  the error condition, if available).}
function WebPIDecode(const Data: PByte; Data_Size: SIZE_T;
  Config: PWebPDecoderConfig): PWebPIDecoder; cdecl; external;

{ Non-incremental version. This version decodes the full data at once, taking
  'config' into account. Returns decoding status (which should be VP8_STATUS_OK
  if the decoding was successful). Note that 'config' cannot be NULL.}
function WebPDecode(const Data: PByte; Data_Size: SIZE_T;
  Config: PWebPDecoderConfig): VP8StatusCode; cdecl; external;

{ From webp\encode.h }

const
  WEBP_ENCODER_ABI_VERSION = $020E;    // MAJOR(8b) + MINOR(8b)

{ Return the encoder's version number, packed in hexadecimal using 8bits for
  each of major/minor/revision. E.g: v2.5.7 is 0x020507.}
function WebPGetEncoderVersion: Integer; cdecl; external;

//------------------------------------------------------------------------------
// One-stop-shop call! No questions asked:

{ Returns the size of the compressed data (pointed to by *output), or 0 if
  an error occurred. The compressed data must be released by the caller
  using the call 'WebPFree(*output)'.
  These functions compress using the lossy format, and the quality_factor
  can go from 0 (smaller output, lower quality) to 100 (best quality,
  larger output).}
function WebPEncodeRGB(const RGB: PByte; Width: Integer; Height: Integer;
  Stride: Integer; Quality_Factor: Single;
  Output: PPByte): SIZE_T; cdecl; external;
function WebPEncodeBGR(const BGR: PByte; Width: Integer; Height: Integer;
  Stride: Integer; Quality_Factor: Single;
  Output: PPByte): SIZE_T; cdecl; external;
function WebPEncodeRGBA(const RGBA: PByte; Width: Integer; Height: Integer;
  Stride: Integer; Quality_Factor: Single;
  Output: PPByte): SIZE_T; cdecl; external;
function WebPEncodeBGRA(const BGRA: PByte; Width: Integer; Height: Integer;
  Stride: Integer; Quality_Factor: Single;
  Output: PPByte): SIZE_T; cdecl; external;

{ These functions are the equivalent of the above, but compressing in a
  lossless manner. Files are usually larger than lossy format, but will
  not suffer any compression loss.}
function WebPEncodeLosslessRGB(const RGB: PByte; Width: Integer;
  Height: Integer; Stride: Integer; Output: PPByte): SIZE_T; cdecl; external;
function WebPEncodeLosslessBGR(const BGR: PByte; Width: Integer;
  Height: Integer; Stride: Integer; Output: PPByte): SIZE_T; cdecl; external;
function WebPEncodeLosslessRGBA(const RGBA: PByte; Width: Integer;
  Height: Integer; Stride: Integer; Output: PPByte): SIZE_T; cdecl; external;
function WebPEncodeLosslessBGRA(const BGRA: PByte; Width: Integer;
  Height: Integer; Stride: Integer; Output: PPByte): SIZE_T; cdecl; external;

//------------------------------------------------------------------------------
// Coding parameters

// Image characteristics hint for the underlying encoder.
type
  WebPImageHint = (
    WEBP_HINT_DEFAULT = 0,  // default preset.
    WEBP_HINT_PICTURE,      // digital picture, like portrait, inner shot
    WEBP_HINT_PHOTO,        // outdoor photograph, with natural lighting
    WEBP_HINT_GRAPH,        // Discrete tone image (graph, map-tile etc).
    WEBP_HINT_LAST
  );

  // Compression parameters.
  WebPConfig = record
    Lossless: Integer;      // Lossless encoding (0=lossy(default), 1=lossless).
    Quality: Single;        // between 0 and 100. For lossy, 0 gives the smallest
                            // size and 100 the largest. For lossless, this
                            // parameter is the amount of effort put into the
                            // compression: 0 is the fastest but gives larger
                            // files compared to the slowest, but best, 100.
    Method: Integer;        // quality/speed trade-off (0=fast, 6=slower-better)

    Image_Hint: WebPImageHint;  // Hint for image type (lossless only for now).

    Target_Size: Integer;   // if non-zero, set the desired target size in bytes.
                            // Takes precedence over the 'compression' parameter.
    Target_PSNR: Single;    // if non-zero, specifies the minimal distortion to
                            // try to achieve. Takes precedence over target_size.
    Segments: Integer;      // maximum number of segments to use, in [1..4]
    SNS_Strength: Integer;  // Spatial Noise Shaping. 0=off, 100=maximum.
    Filter_Strength: Integer;  // range: [0 = off .. 100 = strongest]
    Filter_Sharpness: Integer; // range: [0 = off .. 7 = least sharp]
    Filter_Type: Integer;   // filtering type: 0 = simple, 1 = strong (only used
                            // if filter_strength > 0 or autofilter > 0)
    AutoFilter: Integer;    // Auto adjust filter's strength [0 = off, 1 = on]
    Alpha_Compression: Integer;  // Algorithm for encoding the alpha plane (0 = none,
                                 // 1 = compressed with WebP lossless). Default is 1.
    Alpha_Filtering: Integer;  // Predictive filtering method for alpha plane.
                               //  0: none, 1: fast, 2: best. Default if 1.
    Alpha_Quality: Integer; // Between 0 (smallest size) and 100 (lossless).
                            // Default is 100.
    Pass: Integer;          // number of entropy-analysis passes (in [1..10]).

    Show_Compressed: Integer;  // if true, export the compressed picture back.
                               // In-loop filtering is not applied.
    PreProcessing: Integer; // preprocessing filter:
                            // 0=none, 1=segment-smooth, 2=pseudo-random dithering
    Partitions: Integer;    // log2(number of token partitions) in [0..3]. Default
                            // is set to 0 for easier progressive decoding.
    Partition_Limit: Integer;  // quality degradation allowed to fit the 512k limit
                               // on prediction modes coding (0: no degradation,
                               // 100: maximum possible degradation).
    Emulate_Jpeg_Size: Integer;  // If true, compression parameters will be remapped
                                 // to better match the expected output size from
                                 // JPEG compression. Generally, the output size will
                                 // be similar but the degradation will be lower.
    Thread_Level: Integer;  // If non-zero, try and use multi-threaded encoding.
    Low_Memory: Integer;    // If set, reduce memory usage (but increase CPU use).

    Near_Lossless: Integer; // Near lossless encoding [0 = max loss .. 100 = off
                            // (default)].
    Exact: Integer;         // if non-zero, preserve the exact RGB values under
                            // transparent area. Otherwise, discard this invisible
                            // RGB information for better compression. The default
                            // value is 0.

    Use_Delta_Palette: Integer; // reserved for future lossless feature
    Use_Sharp_YUV: Integer;     // if needed, use sharp (and slow) RGB->YUV conversion

    Pad: array [0..1] of Cardinal;  // padding for later use
  end;
  PWebPConfig = ^WebPConfig;

  { Enumerate some predefined settings for WebPConfig, depending on the type
    of source picture. These presets are used when calling WebPConfigPreset().}
  WebPPreset = (
    WEBP_PRESET_DEFAULT = 0,  // default preset.
    WEBP_PRESET_PICTURE,      // digital picture, like portrait, inner shot
    WEBP_PRESET_PHOTO,        // outdoor photograph, with natural lighting
    WEBP_PRESET_DRAWING,      // hand or line drawing, with high-contrast details
    WEBP_PRESET_ICON,         // small-sized colorful images
    WEBP_PRESET_TEXT          // text-like
  );

{ Internal, version-checked, entry point.}
function WebPConfigInitInternal(Config: PWebPConfig; Preset:
  WebPPreset; Quality: Single; Version: Integer): Integer; cdecl; external;

{ Should always be called, to initialize a fresh WebPConfig structure before
  modification. Returns false in case of version mismatch. WebPConfigInit()
  must have succeeded before using the 'config' object.
  Note that the default values are lossless=0 and quality=75.}
function WebPConfigInit(Config: PWebPConfig): Integer;

{ This function will initialize the configuration according to a predefined
  set of parameters (referred to by 'preset') and a given quality factor.
  This function can be called as a replacement to WebPConfigInit(). Will
  return false in case of error.}
function WebPConfigPreset(Config: PWebPConfig; Preset: WebPPreset;
  Quality: Single): Integer;

{ Activate the lossless compression mode with the desired efficiency level
  between 0 (fastest, lowest compression) and 9 (slower, best compression).
  A good default level is '6', providing a fair tradeoff between compression
  speed and final compressed size.
  This function will overwrite several fields from config: 'method', 'quality'
  and 'lossless'. Returns false in case of parameter error.}
function WebPConfigLosslessPreset(Config: PWebPConfig;
  Level: Integer): Integer; cdecl; external;

{ Returns true if 'config' is non-NULL and all configuration parameters are
  within their valid ranges.}
function WebPValidateConfig(const Config: PWebPConfig): Integer; cdecl; external;

//------------------------------------------------------------------------------
// Input / Output
// Structure for storing auxiliary statistics.

type
  WebPAuxStats = record
    Coded_Size: Integer;         // final size

    PSNR: array [0..4] of Single;           // peak-signal-to-noise ratio for Y/U/V/All/Alpha
    Block_Count: array [0..2] of Integer;   // number of intra4/intra16/skipped macroblocks
    Header_Bytes: array [0..1] of Integer;  // approximate number of bytes spent for header
                                 // and mode-partition #0
    Residual_Bytes: array [0..2] of array [0..3] of Integer; // approximate number of bytes spent for
                                 // DC/AC/uv coefficients for each (0..3) segments.
    Segment_Size: array [0..3] of Integer;  // number of macroblocks in each segments
    Segment_Quant: array [0..3] of Integer; // quantizer values for each segments
    Segment_Level: array [0..3] of Integer; // filtering strength for each segments [0..63]

    Alpha_Data_Size: Integer;    // size of the transparency data
    Layer_Data_Size: Integer;    // size of the enhancement layer data

    // lossless encoder statistics
    Lossless_Features: Cardinal;  // bit0:predictor bit1:cross-color transform
                                  // bit2:subtract-green bit3:color indexing
    Histogram_Bits: Integer;      // number of precision bits of histogram
    Transform_Bits: Integer;      // precision bits for transform
    Cache_Bits: Integer;          // number of bits for color cache lookup
    Palette_Size: Integer;        // number of color in palette, if used
    Lossless_Size: Integer;       // final lossless size
    Lossless_Hdr_Size: Integer;   // lossless header (transform, huffman etc) size
    Lossless_Data_Size: Integer;  // lossless image data size

    Pad: array [0..1] of Cardinal;  // padding for later use
  end;
  PWebPAuxStats = ^WebPAuxStats;

  // Color spaces.
  WebPEncCSP = (
    // chroma sampling
    WEBP_YUV420  = 0,        // 4:2:0
    WEBP_YUV420A = 4,        // alpha channel variant
    WEBP_CSP_UV_MASK = 3,    // bit-mask to get the UV sampling factors
    WEBP_CSP_ALPHA_BIT = 4   // bit that is set if alpha is present
  );

  // Encoding error conditions.
  WebPEncodingError = (
    VP8_ENC_OK = 0,
    VP8_ENC_ERROR_OUT_OF_MEMORY,            // memory error allocating objects
    VP8_ENC_ERROR_BITSTREAM_OUT_OF_MEMORY,  // memory error while flushing bits
    VP8_ENC_ERROR_NULL_PARAMETER,           // a pointer parameter is NULL
    VP8_ENC_ERROR_INVALID_CONFIGURATION,    // configuration is invalid
    VP8_ENC_ERROR_BAD_DIMENSION,            // picture has invalid width/height
    VP8_ENC_ERROR_PARTITION0_OVERFLOW,      // partition is bigger than 512k
    VP8_ENC_ERROR_PARTITION_OVERFLOW,       // partition is bigger than 16M
    VP8_ENC_ERROR_BAD_WRITE,                // error while flushing bytes
    VP8_ENC_ERROR_FILE_TOO_BIG,             // file is bigger than 4G
    VP8_ENC_ERROR_USER_ABORT,               // abort request by user
    VP8_ENC_ERROR_LAST                      // list terminator. always last.
  );

// maximum width/height allowed (inclusive), in pixels
const
  WEBP_MAX_DIMENSION = 16383;

type
  PWebPPicture = ^WebPPicture;

{ Signature for output function. Should return true if writing was successful.
  data/data_size is the segment of data to write, and 'picture' is for
  reference (and so one can make use of picture->custom_ptr).}
  WebPWriterFunction = function(const Data: PByte; Data_Size: SIZE_T; const Picture: PWebPPicture): Integer;
  PWebPWriterFunction = ^WebPWriterFunction;

{ WebPMemoryWrite: a special WebPWriterFunction that writes to memory using
  the following WebPMemoryWriter object (to be set as a custom_ptr).}
  WebPMemoryWriter = record
    Mem: PByte;         // final buffer (of size 'max_size', larger than 'size').
    Size: SIZE_T;       // final size
    Max_Size: SIZE_T;   // total capacity
    Pad: array [0..0] of Cardinal; // padding for later use
  end;
  PWebPMemoryWriter = ^WebPMemoryWriter;

{ Progress hook, called from time to time to report progress. It can return
  false to request an abort of the encoding process, or true otherwise if
  everything is OK.}
  WebPProgressHook = function(Percent: Integer; const Picture: PWebPPicture): Integer;

// Main exchange structure (input samples, output bytes, statistics)
  WebPPicture = record
    //   INPUT
    //////////////
    // Main flag for encoder selecting between ARGB or YUV input.
    // It is recommended to use ARGB input (*argb, argb_stride) for lossless
    // compression, and YUV input (*y, *u, *v, etc.) for lossy compression
    // since these are the respective native colorspace for these formats.
    Use_ARGB: Integer;

    // YUV input (mostly used for input to lossy compression)
    ColorSpace: WebPEncCSP;    // colorspace: should be YUV420 for now (=Y'CbCr).
    Width, Height: Integer;    // dimensions (less or equal to WEBP_MAX_DIMENSION)
    Y, U, V: PByte;            // pointers to luma/chroma planes.
    Y_Stride, UV_Stride: Integer; // luma/chroma strides.
    A: PByte;                  // pointer to the alpha plane
    A_Stride: Integer;         // stride of the alpha plane
    Pad1: array [0..1] of Cardinal; // padding for later use

    // ARGB input (mostly used for input to lossless compression)
    ARGB: PCardinal;           // Pointer to argb (32 bit) plane.
    ARGB_Stride: Integer;      // This is stride in pixels units, not bytes.
    Pad2: array [0..2] of Cardinal; // padding for later use

    //   OUTPUT
    ///////////////
    // Byte-emission hook, to store compressed bytes as they are ready.
    Writer: WebPWriterFunction; // can be NULL
    Custom_Ptr: Pointer;        // can be used by the writer.

    // map for extra information (only for lossy compression mode)
    Extra_Info_Type: Integer; // 1: intra type, 2: segment, 3: quant
                              // 4: intra-16 prediction mode,
                              // 5: chroma prediction mode,
                              // 6: bit cost, 7: distortion
    Extra_Info: PByte;        // if not NULL, points to an array of size
                              // ((width + 15) / 16) * ((height + 15) / 16) that
                              // will be filled with a macroblock map, depending
                              // on extra_info_type.

    //   STATS AND REPORTS
    ///////////////////////////
    // Pointer to side statistics (updated only if not NULL)
    Stats: PWebPAuxStats;

    // Error code for the latest error encountered during encoding
    Error_Code: WebPEncodingError;

    // If not NULL, report progress during encoding.
    Progress_Hook: WebPProgressHook;

    User_Data: PByte;        // this field is free to be set to any value and
                               // used during callbacks (like progress-report e.g.).

    Pad3: array [0..2] of Cardinal;       // padding for later use

    // Unused for now
    Pad4, Pad5: PByte;
    Pad6: array [0..7] of Cardinal;       // padding for later use

    // PRIVATE FIELDS
    ////////////////////
    Memory_: Pointer;          // row chunk of memory for yuva planes
    Memory_ARGB_: Pointer;     // and for argb too.
    Pad7: array [0..1] of Pointer;          // padding for later use
  end;

{ The following must be called first before any use.}
procedure WebPMemoryWriterInit(Writer: PWebPMemoryWriter); cdecl; external;

{ The following must be called to deallocate writer->mem memory. The 'writer'
  object itself is not deallocated.}
procedure WebPMemoryWriterClear(Writer: PWebPMemoryWriter); cdecl; external;

{ The custom writer to be used with WebPMemoryWriter as custom_ptr. Upon
  completion, writer.mem and writer.size will hold the coded data.
  writer.mem must be freed by calling WebPMemoryWriterClear.}
function WebPMemoryWrite(const Data: PByte; Data_Size: SIZE_T;
  const Picture: PWebPPicture): Integer; cdecl; external;

{ Internal, version-checked, entry point}
function WebPPictureInitInternal(Picture: PWebPPicture;
  Version: Integer): Integer; cdecl; external;

{ Should always be called, to initialize the structure. Returns false in case
  of version mismatch. WebPPictureInit() must have succeeded before using the
  'picture' object.
  Note that, by default, use_argb is false and colorspace is WEBP_YUV420.}
function WebPPictureInit(Picture: PWebPPicture): Integer;

//------------------------------------------------------------------------------
// WebPPicture utils

{ Convenience allocation / deallocation based on picture->width/height:
  Allocate y/u/v buffers as per colorspace/width/height specification.
  Note! This function will free the previous buffer if needed.
  Returns false in case of memory error.}
function WebPPictureAlloc(Picture: PWebPPicture): Integer; cdecl; external;

{ Release the memory allocated by WebPPictureAlloc() or WebPPictureImport*().
  Note that this function does _not_ free the memory used by the 'picture'
  object itself.
  Besides memory (which is reclaimed) all other fields of 'picture' are
  preserved.}
procedure WebPPictureFree(Picture: PWebPPicture); cdecl; external;

{ Copy the pixels of *src into *dst, using WebPPictureAlloc. Upon return, *dst
  will fully own the copied pixels (this is not a view). The 'dst' picture need
  not be initialized as its content is overwritten.
  Returns false in case of memory allocation error.}
function WebPPictureCopy(const Src: PWebPPicture;
  Dst: PWebPPicture): Integer; cdecl; external;

{ Compute the single distortion for packed planes of samples.
  'src' will be compared to 'ref', and the raw distortion stored into
  '*distortion'. The refined metric (log(MSE), log(1 - ssim),...' will be
  stored in '*result'.
  'x_step' is the horizontal stride (in bytes) between samples.
  'src/ref_stride' is the byte distance between rows.
  Returns false in case of error (bad parameter, memory allocation error, ...).}
function WebPPlaneDistortion(const Src: PByte; Src_Stride: SIZE_T;
  const Ref: PByte; Ref_Stride: SIZE_T;
  Width: Integer; Height: Integer;
  X_Step: SIZE_T;
  TypeD: Integer; // 0 = PSNR, 1 = SSIM, 2 = LSIM
  Distortion: PFloat; Result: PFloat): Integer; cdecl; external;

type
  DistortionResult = array [0..4] of Float;

{ Compute PSNR, SSIM or LSIM distortion metric between two pictures. Results
  are in dB, stored in result[] in the B/G/R/A/All order. The distortion is
  always performed using ARGB samples. Hence if the input is YUV(A), the
  picture will be internally converted to ARGB (just for the measurement).
  Warning: this function is rather CPU-intensive.}
function WebPPictureDistortion(const Src: PWebPPicture; const Ref: PWebPPicture;
  Metric_Type: Integer;        // 0 = PSNR, 1 = SSIM, 2 = LSIM
  Result: DistortionResult): Integer; cdecl; external;

{ self-crops a picture to the rectangle defined by top/left/width/height.
  Returns false in case of memory allocation error, or if the rectangle is
  outside of the source picture.
  The rectangle for the view is defined by the top-left corner pixel
  coordinates (left, top) as well as its width and height. This rectangle
  must be fully be comprised inside the 'src' source picture. If the source
  picture uses the YUV420 colorspace, the top and left coordinates will be
  snapped to even values.}
function WebPPictureCrop(Picture: PWebPPicture;
  Left: Integer; Top: Integer;
  Width: Integer; Height: Integer): Integer; cdecl; external;

{ Extracts a view from 'src' picture into 'dst'. The rectangle for the view
  is defined by the top-left corner pixel coordinates (left, top) as well
  as its width and height. This rectangle must be fully be comprised inside
  the 'src' source picture. If the source picture uses the YUV420 colorspace,
  the top and left coordinates will be snapped to even values.
  Picture 'src' must out-live 'dst' picture. Self-extraction of view is allowed
  ('src' equal to 'dst') as a mean of fast-cropping (but note that doing so,
  the original dimension will be lost). Picture 'dst' need not be initialized
  with WebPPictureInit() if it is different from 'src', since its content will
  be overwritten.
  Returns false in case of memory allocation error or invalid parameters.}
function WebPPictureView(const Src: PWebPPicture; Left: Integer; Top: Integer;
  Width: Integer; Height: Integer; Dst: PWebPPicture): Integer; cdecl; external;

{ Returns true if the 'picture' is actually a view and therefore does
  not own the memory for pixels.}
function WebPPictureIsView(Picture: PWebPPicture): Integer; cdecl; external;

{ Rescale a picture to new dimension width x height.
  If either 'width' or 'height' (but not both) is 0 the corresponding
  dimension will be calculated preserving the aspect ratio.
  No gamma correction is applied.
  Returns false in case of error (invalid parameter or insufficient memory).}
function WebPPictureRescale(Picture: PWebPPicture;
  Width: Integer; Height: Integer): Integer; cdecl; external;

{ Colorspace conversion function to import RGB samples.
  Previous buffer will be free'd, if any.
  *rgb buffer should have a size of at least height * rgb_stride.
  Returns false in case of memory error.}
function WebPPictureImportRGB(Picture: PWebPPicture; const RGB: PByte;
  RGB_Stride: Integer): Integer; cdecl; external;
{ Same, but for RGBA buffer.}
function WebPPictureImportRGBA(Picture: PWebPPicture; const RGBA: PByte;
  RGBA_Stride: Integer): Integer; cdecl; external;
{ Same, but for RGBA buffer. Imports the RGB direct from the 32-bit format
  input buffer ignoring the alpha channel. Avoids needing to copy the data
  to a temporary 24-bit RGB buffer to import the RGB only.}
function WebPPictureImportRGBX(Picture: PWebPPicture; const RGBX: PByte;
  RGBX_Stride: Integer): Integer; cdecl; external;

{ Variants of the above, but taking BGR(A|X) input.}
function WebPPictureImportBGR(Picture: PWebPPicture; const BGR: PByte;
  BGR_Stride: Integer): Integer; cdecl; external;
function WebPPictureImportBGRA(Picture: PWebPPicture; const BGRA: PByte;
  BGRA_Stride: Integer): Integer; cdecl; external;
function WebPPictureImportBGRX(Picture: PWebPPicture; const BGRX: PByte;
  BGRX_Stride: Integer): Integer; cdecl; external;

{ Converts picture->argb data to the YUV420A format. The 'colorspace'
  parameter is deprecated and should be equal to WEBP_YUV420.
  Upon return, picture->use_argb is set to false. The presence of real
  non-opaque transparent values is detected, and 'colorspace' will be
  adjusted accordingly. Note that this method is lossy.
  Returns false in case of error.}
function WebPPictureARGBToYUVA(Picture: PWebPPicture;
  ColorSpace: WebPEncCSP {colorspace := WEBP_YUV420}): Integer; cdecl; external;

{ Same as WebPPictureARGBToYUVA(), but the conversion is done using
  pseudo-random dithering with a strength 'dithering' between
  0.0 (no dithering) and 1.0 (maximum dithering). This is useful
  for photographic picture.}
function WebPPictureARGBToYUVADithered(Picture: PWebPPicture;
  ColorSpace: WebPEncCSP; Dithering: Float): Integer; cdecl; external;

{ Performs 'sharp' RGBA->YUVA420 downsampling and colorspace conversion.
  Downsampling is handled with extra care in case of color clipping. This
  method is roughly 2x slower than WebPPictureARGBToYUVA() but produces better
  and sharper YUV representation.
  Returns false in case of error.}
function WebPPictureSharpARGBToYUVA(Picture: PWebPPicture): Integer; cdecl; external;
{ kept for backward compatibility:}
function WebPPictureSmartARGBToYUVA(Picture: PWebPPicture): Integer; cdecl; external;

{ Converts picture->yuv to picture->argb and sets picture->use_argb to true.
  The input format must be YUV_420 or YUV_420A. The conversion from YUV420 to
  ARGB incurs a small loss too.
  Note that the use of this colorspace is discouraged if one has access to the
  raw ARGB samples, since using YUV420 is comparatively lossy.
  Returns false in case of error.}
function WebPPictureYUVAToARGB(Picture: PWebPPicture): Integer; cdecl; external;

{ Helper function: given a width x height plane of RGBA or YUV(A) samples
  clean-up or smoothen the YUV or RGB samples under fully transparent area,
  to help compressibility (no guarantee, though).}
procedure WebPCleanupTransparentArea(Picture: PWebPPicture); cdecl; external;

{ Scan the picture 'picture' for the presence of non fully opaque alpha values.
  Returns true in such case. Otherwise returns false (indicating that the
  alpha plane can be ignored altogether e.g.).}
function WebPPictureHasTransparency(const Picture: PWebPPicture): Integer; cdecl; external;

{ Remove the transparency information (if present) by blending the color with
  the background color 'background_rgb' (specified as 24bit RGB triplet).
  After this call, all alpha values are reset to 0xff.}
procedure WebPBlendAlpha(Picture: PWebPPicture; Background_RGB: Cardinal); cdecl; external;

//------------------------------------------------------------------------------
// Main call

{ Main encoding call, after config and picture have been initialized.
  'picture' must be less than 16384x16384 in dimension (cf WEBP_MAX_DIMENSION),
  and the 'config' object must be a valid one.
  Returns false in case of error, true otherwise.
  In case of error, picture->error_code is updated accordingly.
  'picture' can hold the source samples in both YUV(A) or ARGB input, depending
  on the value of 'picture->use_argb'. It is highly recommended to use
  the former for lossy encoding, and the latter for lossless encoding
  (when config.lossless is true). Automatic conversion from one format to
  another is provided but they both incur some loss.}
function WebPEncode(const Config: PWebPConfig; Picture: PWebPPicture): Integer; cdecl; external;

implementation

{$IFDEF FPC}
function c_udivdi3(Num, Den: UInt64): UInt64; cdecl; [public, alias: '___udivdi3'];
begin
  Result := Num div Den;
end;

function c_umoddi3(Num, Den: UInt64): UInt64; cdecl; [public, alias: '___umoddi3'];
begin
  Result := Num mod Den;
end;

function c_divdi3(Num, Den: Int64): Int64; cdecl; [public, alias: '___divdi3'];
begin
  Result := Num div Den;
end;

function c_pow(const Base, Exponent: Double): Double; cdecl; [public, alias: '_pow'];
begin
  if Exponent = 0.0 then
    Result := 1.0
  else if (Base = 0.0) and (Exponent > 0.0) then
    Result := 0.0
  else
    Result := Exp(Exponent * Ln(Base));
end;

function c_log(X: Double): Double; cdecl; [public, alias: '_log'];
begin
  Result := Ln(X);
end;

procedure c_chkstk_ms; cdecl; [public, alias: '___chkstk_ms'];
begin
  // Dummy
end;
{$ENDIF}

function WebPInitDecBuffer(Buffer: PWebPDecBuffer): Integer;
begin
  Result := WebPInitDecBufferInternal(buffer, WEBP_DECODER_ABI_VERSION);
end;

function WebPIDecGetYUV(const IDec: PWebPIDecoder; Last_Y: PInteger; U: PPByte; V: PPByte; Width: PInteger; Height: PInteger; Stride: PInteger; UV_Stride: PInteger): PByte;
begin
  Result := WebPIDecGetYUVA(IDec, Last_Y, U, V, nil, Width, Height, Stride, UV_stride, nil);
end;

function WebPGetFeatures(const Data: PByte; Data_Size: SIZE_T; Features: PWebPBitstreamFeatures): VP8StatusCode;
begin
  Result := WebPGetFeaturesInternal(Data, Data_Size, Features, WEBP_DECODER_ABI_VERSION);
end;

function WebPInitDecoderConfig(Config: PWebPDecoderConfig): Integer;
begin
  Result := WebPInitDecoderConfigInternal(Config, WEBP_DECODER_ABI_VERSION);
end;

function WebPConfigInit(Config: PWebPConfig): Integer;
begin
  Result := WebPConfigInitInternal(Config, WEBP_PRESET_DEFAULT, 75, WEBP_ENCODER_ABI_VERSION);
end;

function WebPConfigPreset(Config: PWebPConfig; Preset: WebPPreset; Quality: Single): Integer;
begin
  Result := WebPConfigInitInternal(Config, Preset, Quality, WEBP_ENCODER_ABI_VERSION);
end;

function WebPPictureInit(Picture: PWebPPicture): Integer;
begin
  Result := WebPPictureInitInternal(Picture, WEBP_ENCODER_ABI_VERSION);
end;

{$IF Defined(MSWINDOWS)}
  {$IF Defined(DCC)}
    { Delphi Win32 }
    { Link object files created with C++ Builder.}

    const
      { MS C Runtime library needed for importing std C functions.}
      MSCRuntimeLib = 'msvcrt.dll';


    procedure _llmul; cdecl;
    asm
      { from Delphi's System.pas __llmul }
      push  edx
      push  eax

      mov   eax, [esp+16]
      mul   dword ptr [esp]
      mov   ecx, eax

      mov   eax, [esp+4]
      mul   dword ptr [esp+12]
      add   ecx, eax

      mov   eax, [esp]
      mul   dword ptr [esp+12]
      add   edx, ecx

      pop   ecx
      pop   ecx

      ret     8
    end;

    function pow(const Base, Exponent: Double): Double; cdecl;
    begin
      if Exponent = 0.0 then
        Result := 1.0
      else if (Base = 0.0) and (Exponent > 0.0) then
        Result := 0.0
      else
        Result := Exp(Exponent * Ln(Base));
    end;

    { C library imports }
    function malloc(size: Cardinal): Pointer; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_malloc'{$ENDIF};
    function calloc(nelem, elsize: Cardinal): Pointer; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_calloc'{$ENDIF};
    procedure free(ptr: Pointer); cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_free'{$ENDIF};
    function realloc(ptr: Pointer; size: Cardinal): Pointer; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_realloc'{$ENDIF};
    function memset(s: Pointer; c, n: Cardinal): Pointer; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_memset'{$ENDIF};
    function memcpy(s1, s2: Pointer; n: Cardinal): Pointer; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_memcpy'{$ENDIF};
    function floor(const x: Double): Double; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_floor'{$ENDIF};
    function ceil(const num: Double): Double; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_ceil'{$ENDIF};
    function printf(format: PAnsiChar): Integer; cdecl; varargs; external MSCRuntimeLib{$IFDEF BCB} name '_printf'{$ENDIF};
    function fprintf(f: Pointer; format: PAnsiChar): Integer; cdecl; varargs; external MSCRuntimeLib{$IFDEF BCB} name '_fprintf'{$ENDIF};
    function vsprintf(s, format: PAnsiChar): Integer; cdecl; varargs; external MSCRuntimeLib{$IFDEF BCB} name '_vsprintf'{$ENDIF};
    function _ftol(x: Single): LongInt; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '__ftol'{$ENDIF};
    function strcpy(s1, s2: PAnsiChar): PAnsiChar; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_strcpy'{$ENDIF};
    function wcscpy(s1, s2: PAnsiChar): PAnsiChar; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_wstrcpy'{$ENDIF};
    function strncpy(s1, s2: PAnsiChar; maxlen: Integer): PAnsiChar; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_strncpy'{$ENDIF};
    function strlen(s: PAnsiChar): Integer; cdecl; external MSCRuntimeLib{$IFDEF BCB} name '_strlen'{$ENDIF};
  {$ELSEIF Defined(FPC)}
    { Free Pascal Win32 }
    { Link WebP codec static library and C runtime library.}
    {$LINKLIB libwebpwin32.a}
    {$LINKLIB libmsvcrt.a}
    {$LINKLIB libkernel32.a}
  {$IFEND}
{$IFEND}

end.
